"""
A helper library for Django to work with birthdays in models.

"""
from django.core.exceptions import FieldError
from django.db.models.fields import PositiveSmallIntegerField, DateField
from django.db.models.signals import pre_save


def pre_save_listener(instance, **kwargs):
    field = instance._meta.birthday_field
    birthday = getattr(instance, field.name)
    if not birthday:
        return
    doy = birthday.timetuple().tm_yday
    setattr(instance, field.doy_name, doy)


# An un-editable internal field that will hold the cooresponding day-of-the-
# month for the given date.
internal_field = PositiveSmallIntegerField(editable=False, default=None, null=True)


class BirthdayField(DateField):
    def contribute_to_class(self, cls, name):
        super(BirthdayField, self).contribute_to_class(cls, name)
        if hasattr(cls._meta, 'birthday_field'):
            raise FieldError('Multiple BirthdayFields are not supported on a single model.')
        cls._meta.birthday_field = self
        self.doy_name = '%s_dayofyear' % name
        internal_field.contribute_to_class(cls, self.doy_name)
        pre_save.connect(pre_save_listener, sender=cls)



from django import forms
from django.core import exceptions
from django.db import models, router
from django.db.models import Field, ForeignKey

from south.modelsinspector import add_introspection_rules


class CharField(Field):
    def formfield(self, **kwargs):
        defaults = {'widget': forms.TextInput}
        defaults.update(kwargs)
        return super(CharField, self).formfield(**defaults)

    def db_type(self, connection):
        if self.max_length:
            return 'varchar(%s)' % self.max_length
        return 'text'


class EmailField(CharField):
    def formfield(self, **kwargs):
        defaults = {'form_class': forms.EmailField}
        defaults.update(kwargs)
        return super(EmailField, self).formfield(**defaults)


class SlugField(CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('db_index', True)
        super(SlugField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        defaults = {'form_class': forms.SlugField}
        defaults.update(kwargs)
        return super(SlugField, self).formfield(**defaults)


class URLField(CharField):
    def __init__(self, verbose_name=None, name=None, verify_exists=True, **kwargs):
        self.verify_exists = verify_exists
        super(URLField, self).__init__(verbose_name, name, **kwargs)

    def formfield(self, **kwargs):
        defaults = {'form_class': forms.URLField, 'verify_exists': self.verify_exists}
        defaults.update(kwargs)
        return super(URLField, self).formfield(**defaults)


class CustomManagerForeignKey(ForeignKey):
    """
    Inspired by:
    <http://www.hoboes.com/Mimsy/hacks/custom-managers-django-foreignkeys/>

    """
    def __init__(self, *args, **kwargs):
        if 'manager' in kwargs:
            self.custom_manager = kwargs['manager']()
            del kwargs['manager']
        else:
            self.custom_manager = None
        super(CustomManagerForeignKey, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        field = super(CustomManagerForeignKey, self).formfield(**kwargs)
        if self.custom_manager:
            field.queryset = self.custom_manager
        return field

    def validate(self, value, model_instance):
        if self.rel.parent_link:
            return
        super(models.ForeignKey, self).validate(value, model_instance)
        if value is None:
            return
        if self.custom_manager:
            manager = self.custom_manager
        else:
            using = router.db_for_read(model_instance.__class__, instance=model_instance)
            manager = self.rel.to._default_manager.using(using)
        qs = manager.filter(**{self.rel.field_name: value})
        qs = qs.complex_filter(self.rel.limit_choices_to)
        if not qs.exists():
            raise exceptions.ValidationError(
                self.error_messages['invalid'] % {
                    'model': self.rel.to._meta.verbose_name,
                    'pk': value
                }
            )


# South custom field introspection rules.
add_introspection_rules([], [r'^ohashi\.db\.fields\.CustomManagerForeignKey'])



from datetime import date

from django.db.models.manager import Manager
from django.db.models.query_utils import Q


CASE = "CASE WHEN %(bdoy)s<%(cdoy)s THEN %(bdoy)s+365 ELSE %(bdoy)s END"


def _order(manager, reverse=False, case=False):
    cdoy = date.today().timetuple().tm_yday
    bdoy = manager._birthday_doy_field
    doys = {'cdoy': cdoy, 'bdoy': bdoy}
    if case:
        qs = manager.extra(select={'internal_bday_order': CASE % doys})
        order_field = 'internal_bday_order'
    else:
        qs = manager.all()
        order_field = bdoy
    if reverse:
        return qs.order_by('-%s' % order_field)
    return qs.order_by('%s' % order_field)


class BirthdayManager(Manager):
    @property
    def _birthday_doy_field(self):
        return self.model._meta.birthday_field.doy_name

    def _doy(self, day):
        if not day:
            day = date.today()
        return day.timetuple().tm_yday

    def get_upcoming_birthdays(self, days=30, after=None, include_day=True, order=True, reverse=False):
        today = self._doy(after)
        limit = today + days
        q = Q(**{'%s__gt%s' % (self._birthday_doy_field, 'e' if include_day else ''): today})
        q &= Q(**{'%s__lt' % self._birthday_doy_field: limit})
        if limit > 365:
            limit = limit - 365
            today = 1
            q2 = Q(**{'%s__gte' % self._birthday_doy_field: today})
            q2 &= Q(**{'%s__lt' % self._birthday_doy_field: limit})
            q = q | q2
        if order:
            qs = _order(self, reverse, True)
            return qs.filter(q)
        else:
            return self.filter(q)

    def get_birthdays(self, day=None):
        today = self._doy(day)
        return self.filter(**{self._birthday_doy_field: today})

    def order_by_birthday(self, reverse=False):
        return _order(self, reverse)
